# fslib.py
# Library to support filesystem classes.
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#                    David Cantrell <dcantrell@redhat.com>
#                    Anne Mulhern <amulhern@redhat.com>

kernel_filesystems = []
nodev_filesystems = []


def update_kernel_filesystems():
    with open("/proc/filesystems") as filesystems:
        for line in filesystems:
            fields = line.split()
            fstype = fields[-1]
            kernel_filesystems.append(fstype)
            if fields[0] == "nodev":
                nodev_filesystems.append(fstype)


update_kernel_filesystems()



# devices/btrfs.py
#
# Copyright (C) 2009-2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import os
import copy
import tempfile

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

from ..devicelibs import btrfs
from ..devicelibs import raid

from .. import errors
from ..flags import flags
from ..storage_log import log_method_call
from .. import udev
from .. import util
from ..formats import get_format, DeviceFormat
from ..size import Size
from ..mounts import mounts_cache

import logging
log = logging.getLogger("blivet")

from .storage import StorageDevice
from .container import ContainerDevice
from .raid import RaidDevice

from ..tasks import availability

from contextlib import contextmanager


class BTRFSDevice(StorageDevice):

    """ Base class for BTRFS volume and sub-volume devices. """
    _type = "btrfs"
    _packages = ["btrfs-progs"]
    _external_dependencies = [availability.BLOCKDEV_BTRFS_PLUGIN]

    def __init__(self, *args, **kwargs):
        """ Passing None or no name means auto-generate one like btrfs.%d """
        if not args or not args[0]:
            args = ("btrfs.%d" % self.id,)

        if kwargs.get("parents") is None:
            raise errors.BTRFSValueError("BTRFSDevice must have at least one parent")

        self.req_size = kwargs.pop("size", None)
        super(BTRFSDevice, self).__init__(*args, **kwargs)

    def update_sysfs_path(self):
        """ Update this device's sysfs path. """
        log_method_call(self, self.name, status=self.status)
        self.parents[0].update_sysfs_path()
        self.sysfs_path = self.parents[0].sysfs_path
        log.debug("%s sysfs_path set to %s", self.name, self.sysfs_path)

    def update_size(self, newsize=None):
        pass

    def _post_create(self):
        super(BTRFSDevice, self)._post_create()
        self.format.exists = True
        self.format.device = self.path

    def _pre_destroy(self):
        """ Preparation and precondition checking for device destruction. """
        super(BTRFSDevice, self)._pre_destroy()
        self.setup_parents(orig=True)

    def _get_size(self):
        return sum((d.size for d in self.parents), Size(0))

    def _set_size(self, newsize):
        raise RuntimeError("cannot directly set size of btrfs volume")

    @property
    def current_size(self):
        return self.size

    @property
    def status(self):
        return self.exists and all(d.status for d in self.parents)

    @property
    def _temp_dir_prefix(self):
        return "btrfs-tmp.%s" % self.id

    @contextmanager
    def _do_temp_mount(self, orig=False):
        if not self.exists:
            raise errors.FSError("format doesn't exist")

        if orig:
            fmt = self.original_format
        else:
            fmt = self.format

        if fmt.status:
            yield fmt.system_mountpoint

        else:
            tmpdir = tempfile.mkdtemp(prefix=self._temp_dir_prefix)
            try:
                util.mount(device=fmt.device, mountpoint=tmpdir, fstype=fmt.type,
                           options=fmt.mountopts)
            except errors.FSError as e:
                log.debug("btrfs temp mount failed: %s", e)
                raise

            try:
                yield tmpdir
            finally:
                util.umount(mountpoint=tmpdir)
                os.rmdir(tmpdir)

    @property
    def path(self):
        return self.parents[0].path if self.parents else None

    @property
    def direct(self):
        """ Is this device directly accessible? """
        return True

    @property
    def fstab_spec(self):
        if self.format.vol_uuid:
            spec = "UUID=%s" % self.format.vol_uuid
        else:
            spec = super(BTRFSDevice, self).fstab_spec
        return spec

    def is_name_valid(self, name):
        # Override StorageDevice.is_name_valid to allow pretty much anything
        return btrfs.is_btrfs_name_valid(name)


class BTRFSVolumeDevice(BTRFSDevice, ContainerDevice, RaidDevice):
    _type = "btrfs volume"
    vol_id = btrfs.MAIN_VOLUME_ID
    _format_class_name = property(lambda s: "btrfs")
    _format_uuid_attr = property(lambda s: "vol_uuid")

    def __init__(self, *args, **kwargs):
        """
            :param str name: the volume name
            :keyword bool exists: does this device exist?
            :keyword :class:`~.size.Size` size: the device's size
            :keyword :class:`~.ParentList` parents: a list of parent devices
            :keyword fmt: this device's formatting
            :type fmt: :class:`~.formats.DeviceFormat`
            :keyword str uuid: UUID of top-level filesystem/volume
            :keyword str sysfs_path: sysfs device path
            :keyword data_level: RAID level for data
            :type data_level: any valid raid level descriptor
            :keyword metadata_level: RAID level for metadata
            :type metadata_level: any valid raid level descriptor
        """
        # pop these arguments before the constructor call to avoid
        # unrecognized keyword error in superclass constructor
        data_level = kwargs.pop("data_level", None)
        metadata_level = kwargs.pop("metadata_level", None)
        create_options = kwargs.pop("create_options", None)

        super(BTRFSVolumeDevice, self).__init__(*args, **kwargs)

        # avoid attribute-defined-outside-init pylint warning
        self._data_level = self._metadata_level = None

        # assign after constructor to avoid AttributeErrors in setter functions
        try:
            self.data_level = data_level
            self.metadata_level = metadata_level
        except errors.BTRFSValueError as e:
            # Could not set the levels, so set loose the parents that were
            # added in superclass constructor.
            for dev in self.parents:
                dev.remove_child(self)
            raise e

        self.subvolumes = []
        self.size_policy = self.size

        if self.parents and not self.format.type:
            label = getattr(self.parents[0].format, "label", None)
            self.format = get_format("btrfs",
                                     exists=self.exists,
                                     label=label,
                                     vol_uuid=self.uuid,
                                     device=self.path,
                                     subvolspec=self.vol_id,
                                     mountopts="subvolid=%d" % self.vol_id,
                                     create_options=create_options)
            self.original_format = copy.deepcopy(self.format)

        self._default_subvolume_id = None

    @property
    def members(self):
        return list(self.parents)

    def _set_level(self, value, data):
        """ Sets a valid level for this device and level type.

            :param object value: value for this RAID level
            :param bool data: True if for data, False if for metadata

            :returns: a valid level for value, if any, else None
            :rtype: :class:`~.devicelibs.raid.RAIDLevel` or NoneType

            :raises :class:`~.errors.BTRFSValueError`: if value represents
            an invalid level.
        """
        level = None
        if value:
            levels = btrfs.raid_levels if data else btrfs.metadata_levels
            try:
                level = self._get_level(value, levels)
            except ValueError as e:
                raise errors.BTRFSValueError(e)

        if data:
            self._data_level = level
        else:
            self._metadata_level = level

    @property
    def data_level(self):
        """ Return the RAID level for data.

            :returns: raid level
            :rtype: an object that represents a raid level
        """
        return self._data_level

    @data_level.setter
    def data_level(self, value):
        """ Set the RAID level for data.

            :param object value: new raid level
            :returns: None
            :raises :class:`~.errors.BTRFSValueError`: if value represents
            an invalid level.
        """
        self._set_level(value, True)

    @property
    def metadata_level(self):
        """ Return the RAID level for metadata.

            :returns: raid level
            :rtype: an object that represents a raid level
        """
        return self._metadata_level

    @metadata_level.setter
    def metadata_level(self, value):
        """ Set the RAID level for metadata.

            :param object value: new raid level
            :returns: None
            :raises :class:`~.errors.BTRFSValueError`: if value represents
            an invalid level.
        """
        self._set_level(value, False)

    @property
    def format_immutable(self):
        return super(BTRFSVolumeDevice, self).format_immutable or self.exists

    def _set_name(self, value):
        self._name = value  # name is not used outside of blivet

    def _set_format(self, fmt):
        """ Set the Device's format. """
        super(BTRFSVolumeDevice, self)._set_format(fmt)
        self.name = "btrfs.%d" % self.id
        label = getattr(self.format, "label", None)
        if label:
            self.name = label

        if not self.exists:
            # propagate mount options specified for members via kickstart
            self.format.mountopts = self.parents[0].format.mountopts

    def _get_size(self):
        # Calculate the size as if it were a RAID with no superblock and a chunk_size of 1
        data_level = self.data_level or raid.Single
        return data_level.get_size([d.size for d in self.parents],
                                   chunk_size=Size(1),
                                   superblock_size_func=lambda x: 0)

    def _remove_parent(self, parent):
        levels = (l for l in (self.data_level, self.metadata_level) if l)
        for l in levels:
            error_msg = self._validate_parent_removal(l, parent)
            if error_msg:
                raise errors.DeviceError(error_msg)
        super(BTRFSVolumeDevice, self)._remove_parent(parent)

    def _add_subvolume(self, vol):
        if vol.name in [v.name for v in self.subvolumes]:
            raise errors.BTRFSValueError("subvolume %s already exists" % vol.name)

        self.subvolumes.append(vol)

    def _remove_subvolume(self, name):
        if name not in [v.name for v in self.subvolumes]:
            raise errors.BTRFSValueError("cannot remove non-existent subvolume %s" % name)

        names = [v.name for v in self.subvolumes]
        self.subvolumes.pop(names.index(name))

    def _get_any_btrfs_mountpoint(self):
        """ Get any of the mountpoints for this btrfs volume.
            This includes mountpoints of subvolumes. The idea is
            to get a mountpoint usable for calling btrfs functions
            like btrfs.list_subvolumes where any mountpoint works.
        """
        # first just check whether this volume is mounted
        if self.format.system_mountpoint:
            return self.format.system_mountpoint
        if self.original_format.system_mountpoint:
            return self.original_format.system_mountpoint

        # now try every possible mountpoint with any subvolspec in our cache
        parents = [p.name for p in self.parents]
        mount_spec = next(((dev, subvol) for dev, subvol in mounts_cache.mountpoints if dev in parents), None)
        if mount_spec:
            try:
                return mounts_cache.get_mountpoints(devspec=mount_spec[0],
                                                    subvolspec=mount_spec[1])[-1]
            except IndexError:
                return None
        return None

    def _list_subvolumes(self, mountpoint, snapshots_only=False):
        subvols = []
        try:
            subvols = blockdev.btrfs.list_subvolumes(mountpoint,
                                                     snapshots_only=snapshots_only)
        except (blockdev.BtrfsError, blockdev.BlockDevNotImplementedError) as e:
            log.debug("failed to list subvolumes: %s", e)
        else:
            self._get_default_subvolume_id()

        return subvols

    def list_subvolumes(self, snapshots_only=False):
        subvols = []
        if flags.auto_dev_updates:
            self.setup(orig=True)

        # for list_subvolumes we can use mountpoint of a subvolume too, the volume
        # itself doesn't need to be mounted
        mountpoint = self._get_any_btrfs_mountpoint()
        if mountpoint:
            return self._list_subvolumes(mountpoint=mountpoint,
                                         snapshots_only=snapshots_only)

        # flags.auto_dev_updates is set --> we'll do a temp mount to get the subvolumes
        if flags.auto_dev_updates:
            try:
                with self._do_temp_mount(orig=True) as mountpoint:
                    subvols = self._list_subvolumes(mountpoint=mountpoint,
                                                    snapshots_only=snapshots_only)
            except errors.FSError:
                pass

        return subvols

    def remove_subvolume(self, name):
        raise NotImplementedError()

    def _get_default_subvolume_id(self):
        subvolid = None
        with self._do_temp_mount() as mountpoint:
            try:
                subvolid = blockdev.btrfs.get_default_subvolume_id(mountpoint)
            except (blockdev.BtrfsError, blockdev.BlockDevNotImplementedError) as e:
                log.debug("failed to get default subvolume id: %s", e)

        self._default_subvolume_id = subvolid

    def _set_default_subvolume_id(self, vol_id):
        """ Set a new default subvolume by id.

            This writes the change to the filesystem, which must be mounted.
        """
        with self._do_temp_mount() as mountpoint:
            try:
                blockdev.btrfs.set_default_subvolume(mountpoint, vol_id)
            except (blockdev.BtrfsError, blockdev.BlockDevNotImplementedError) as e:
                log.error("failed to set new default subvolume id (%s): %s",
                          vol_id, e)
                # The only time we set a new default subvolume is so we can remove
                # the current default. If we can't change the default, we won't be
                # able to remove the subvolume.
                raise
            else:
                self._default_subvolume_id = vol_id

    @property
    def default_subvolume(self):
        default = None
        if self._default_subvolume_id is None:
            return None

        if self._default_subvolume_id == self.vol_id:
            return self

        for sv in self.subvolumes:
            if sv.vol_id == self._default_subvolume_id:
                default = sv
                break

        return default

    def _pre_create(self):
        if any(p.size < btrfs.MIN_MEMBER_SIZE for p in self.parents):
            raise errors.DeviceCreateError("All BTRFS member devices must have size at least %s." % btrfs.MIN_MEMBER_SIZE)

        super(BTRFSVolumeDevice, self)._pre_create()

    def _create(self):
        log_method_call(self, self.name, status=self.status)
        if self.data_level:
            data_level = str(self.data_level)
        else:
            data_level = None
        if self.metadata_level:
            md_level = str(self.metadata_level)
        else:
            md_level = None
        blockdev.btrfs.create_volume([d.path for d in self.parents],
                                     label=self.format.label,
                                     data_level=data_level,
                                     md_level=md_level)

    def _post_create(self):
        super(BTRFSVolumeDevice, self)._post_create()
        info = udev.get_device(self.sysfs_path)
        if not info:
            log.error("failed to get updated udev info for new btrfs volume")
        else:
            self.format.vol_uuid = udev.device_get_uuid(info)

        if not self.format.vol_uuid:
            try:
                bd_info = blockdev.btrfs.filesystem_info(self.parents[0].path)
            except (blockdev.BtrfsError, blockdev.BlockDevNotImplementedError) as e:
                log.error("failed to get filesystem info for new btrfs volume %s", e)
            else:
                self.format.vol_uuid = bd_info.uuid

        self.format.exists = True
        self.original_format.exists = True

    def _destroy(self):
        log_method_call(self, self.name, status=self.status)
        for device in self.parents:
            device.setup(orig=True)
            DeviceFormat(device=device.path, exists=True).destroy()

    def _remove(self, member):
        log_method_call(self, self.name, status=self.status)

        with self._do_temp_mount() as mountpoint:
            blockdev.btrfs.remove_device(mountpoint, member.path)

    def _add(self, member):

        with self._do_temp_mount() as mountpoint:
            blockdev.btrfs.add_device(mountpoint, member.path)

    def populate_ksdata(self, data):
        super(BTRFSVolumeDevice, self).populate_ksdata(data)
        data.data_level = self.data_level.name if self.data_level else None
        data.metadata_level = self.metadata_level.name if self.metadata_level else None
        data.devices = ["btrfs.%d" % p.id for p in self.parents]
        data.preexist = self.exists


class BTRFSSubVolumeDevice(BTRFSDevice):

    """ A btrfs subvolume pseudo-device. """
    _type = "btrfs subvolume"
    _format_immutable = True

    def __init__(self, *args, **kwargs):
        """
            :param str name: the subvolume name
            :keyword bool exists: does this device exist?
            :keyword :class:`~.size.Size` size: the device's size
            :keyword :class:`~.ParentList` parents: a list of parent devices
            :keyword fmt: this device's formatting
            :type fmt: :class:`~.formats.DeviceFormat`
            :keyword str sysfs_path: sysfs device path
        """
        self.vol_id = kwargs.pop("vol_id", None)

        super(BTRFSSubVolumeDevice, self).__init__(*args, **kwargs)

        if len(self.parents) != 1:
            raise errors.BTRFSValueError("%s must have exactly one parent." % self.type)

        if not isinstance(self.parents[0], BTRFSDevice):
            raise errors.BTRFSValueError("%s unique parent must be a BTRFSDevice." % self.type)

        self.volume._add_subvolume(self)

    def _set_format(self, fmt):
        """ Set the Device's format. """
        super(BTRFSSubVolumeDevice, self)._set_format(fmt)
        if self.exists:
            return

        # propagate mount options specified for members via kickstart
        opts = "subvol=%s" % self.name
        has_compress = False
        if self.volume.format.mountopts:
            for opt in self.volume.format.mountopts.split(","):
                # do not add members subvol spec
                if not opt.startswith("subvol"):
                    opts += ",%s" % opt
                if opt.startswith("compress"):
                    has_compress = True

        # add default compression settings
        if flags.btrfs_compression and not has_compress:
            opts += ",compress=%s" % flags.btrfs_compression

        self.format.mountopts = opts

    @property
    def volume(self):
        """Return the first ancestor that is not a BTRFSSubVolumeDevice.

           Note: Assumes that each ancestor in traversal has only one parent.

           Raises a DeviceError if the ancestor found is not a
           BTRFSVolumeDevice.
        """
        parent = self.parents[0]
        vol = None
        while True:
            if not isinstance(parent, BTRFSSubVolumeDevice):
                vol = parent
                break

            parent = parent.parents[0]

        if not isinstance(vol, BTRFSVolumeDevice):
            raise errors.DeviceError("%s %s's first non subvolume ancestor must be a btrfs volume" % (self.type, self.name))
        return vol

    @property
    def container(self):
        return self.volume

    def setup_parents(self, orig=False):
        """ Run setup method of all parent devices. """
        log_method_call(self, name=self.name, orig=orig)
        self.volume.setup(orig=orig)

    def _create(self):
        log_method_call(self, self.name, status=self.status)

        with self.volume._do_temp_mount() as mountpoint:
            blockdev.btrfs.create_subvolume(mountpoint, self.name)

    def _post_create(self):
        super(BTRFSSubVolumeDevice, self)._post_create()
        self.format.vol_uuid = self.volume.format.vol_uuid

    def _destroy(self):
        log_method_call(self, self.name, status=self.status)

        with self.volume._do_temp_mount() as mountpoint:
            if self.volume._default_subvolume_id == self.vol_id:
                # btrfs does not allow removal of the default subvolume
                self.volume._set_default_subvolume_id(self.volume.vol_id)

            blockdev.btrfs.delete_subvolume(mountpoint, self.name)

    def remove_hook(self, modparent=True):
        if modparent:
            self.volume._remove_subvolume(self.name)

        super(BTRFSSubVolumeDevice, self).remove_hook(modparent=modparent)

    def add_hook(self, new=True):
        super(BTRFSSubVolumeDevice, self).add_hook(new=new)
        if new:
            return

        if self not in self.volume.subvolumes:
            self.volume._add_subvolume(self)

    def populate_ksdata(self, data):
        super(BTRFSSubVolumeDevice, self).populate_ksdata(data)
        data.subvol = True
        data.name = self.name
        data.preexist = self.exists

        # Identify the volume this subvolume belongs to by means of its
        # label. If the volume has no label, do nothing.
        # Note that doing nothing will create an invalid kickstart.
        # See rhbz#1072060
        label = self.parents[0].format.label
        if label:
            data.devices = ["LABEL=%s" % label]


class BTRFSSnapShotDevice(BTRFSSubVolumeDevice):

    """ A btrfs snapshot pseudo-device.

        BTRFS snapshots are a specialized type of subvolume that contains a
        source attribute which identifies which subvolume the snapshot was taken
        from. They do not have to be removed when removing the source subvolume.
    """
    _type = "btrfs snapshot"

    def __init__(self, *args, **kwargs):
        """
            :param str name: the subvolume name
            :keyword bool exists: does this device exist?
            :keyword :class:`~.size.Size` size: the device's size
            :keyword :class:`~.ParentList` parents: a list of parent devices
            :keyword fmt: this device's formatting
            :type fmt: :class:`~.formats.DeviceFormat`
            :keyword str sysfs_path: sysfs device path
            :keyword :class:`~.BTRFSDevice` source: the snapshot source
            :keyword bool read_only: create a read-only snapshot

            Snapshot source can be either a subvolume or a top-level volume.

        """
        source = kwargs.pop("source", None)
        if not kwargs.get("exists") and not source:
            # it is possible to remove a source subvol and keep snapshots of it
            raise errors.BTRFSValueError("non-existent btrfs snapshots must have a source")

        if source and not isinstance(source, BTRFSDevice):
            raise errors.BTRFSValueError("btrfs snapshot source must be a btrfs subvolume")

        if source and not source.exists:
            raise errors.BTRFSValueError("btrfs snapshot source must already exist")

        self.source = source
        """ the snapshot's source subvolume """

        self.read_only = kwargs.pop("read_only", False)

        super(BTRFSSnapShotDevice, self).__init__(*args, **kwargs)

        if source and getattr(source, "volume", source) != self.volume:
            self.volume._remove_subvolume(self.name)
            self.parents = []
            raise errors.BTRFSValueError("btrfs snapshot and source must be in the same volume")

    def _create(self):
        log_method_call(self, self.name, status=self.status)

        with self.volume._do_temp_mount() as mountpoint:
            if isinstance(self.source, BTRFSVolumeDevice):
                source_path = mountpoint
            else:
                source_path = "%s/%s" % (mountpoint, self.source.name)

            dest_path = "%s/%s" % (mountpoint, self.name)

            blockdev.btrfs.create_snapshot(source_path, dest_path, ro=self.read_only)

    def depends_on(self, dep):
        return (dep == self.source or
                super(BTRFSSnapShotDevice, self).depends_on(dep))



#
# Copyright (C) 2016  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#
import dbus

from .constants import BUS_NAME, OBJECT_MANAGER_INTERFACE, OBJECT_MANAGER_PATH


class ObjectManager(dbus.service.Object):
    """ Class to implement org.freedesktop.DBus.ObjectManager interface.

        Blivet's ObjectManager interface will manage subtrees for objects that
        variously (and with mutual-exclusivity) implement blivet's Device,
        Format, Action interfaces.
    """
    def __init__(self):
        self._objects = list()
        self._by_id = dict()
        self._by_path = dict()
        super(ObjectManager, self).__init__(bus_name=dbus.service.BusName(BUS_NAME, dbus.SystemBus()),
                                            object_path=OBJECT_MANAGER_PATH)

    @dbus.service.method(dbus_interface=OBJECT_MANAGER_INTERFACE, out_signature='a{oa{sa{sv}}}')
    def GetManagedObjects(self):
        return dict((obj.object_path, {obj.interface: obj.properties}) for obj in self._objects)

    def remove_object(self, obj):
        self._objects.remove(obj)
        del self._by_id[obj.id]
        del self._by_path[obj.object_path]
        self.InterfacesRemoved(obj.object_path, [obj.interface])
        obj.remove_from_connection()

    def add_object(self, obj):
        self._objects.append(obj)
        self._by_id[obj.id] = obj
        self._by_path[obj.object_path] = obj
        self.InterfacesAdded(obj.object_path, {obj.interface: obj.properties})

    @property
    def objects(self):
        return self._objects

    def get_object_by_id(self, obj_id):
        return self._by_id.get(obj_id)

    def get_object_by_path(self, obj_path):
        return self._by_path.get(obj_path)

    @dbus.service.signal(dbus_interface=OBJECT_MANAGER_INTERFACE, signature='oa{sa{sv}}')
    def InterfacesAdded(self, object_path, ifaces_props_dict):
        pass

    @dbus.service.signal(dbus_interface=OBJECT_MANAGER_INTERFACE, signature='oas')
    def InterfacesRemoved(self, object_path, interfaces):
        pass



# dfresize.py
# DeviceFormat resizing classes.
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>

import abc

from six import add_metaclass

from . import task


@add_metaclass(abc.ABCMeta)
class DFResizeTask(task.Task):
    """ The abstract properties that any resize task must have. """

    unit = abc.abstractproperty(doc="Resize unit.")


class UnimplementedDFResize(task.UnimplementedTask, DFResizeTask):

    def __init__(self, a_df):
        """ Initializer.

            :param DeviceFormat a_df: a device format object
        """
        self.df = a_df

    @property
    def unit(self):
        raise NotImplementedError()



import copy
import functools
import glob
import itertools
import os
import subprocess
import re
import sys
import tempfile
import uuid
import hashlib
import warnings
import abc
from decimal import Decimal
from contextlib import contextmanager
from functools import wraps
from collections import namedtuple
from enum import Enum

from .errors import DependencyError
from . import safe_dbus

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

import six

import logging
log = logging.getLogger("blivet")
program_log = logging.getLogger("program")
testdata_log = logging.getLogger("testdata")
console_log = logging.getLogger("blivet.console")

from threading import Lock
# this will get set to anaconda's program_log_lock in enable_installer_mode
program_log_lock = Lock()

try:
    import selinux
except ImportError:
    HAVE_SELINUX = False
else:
    HAVE_SELINUX = True


SYSTEMD_SERVICE = "org.freedesktop.systemd1"
SYSTEMD_MANAGER_PATH = "/org/freedesktop/systemd1"
SYSTEMD_MANAGER_IFACE = "org.freedesktop.systemd1.Manager"
VIRT_PROP_NAME = "Virtualization"


class Path(str):

    """ Path(path, root=None) provides a filesystem path object, which
        automatically normalizes slashes, assumes appends are what you
        always hoped os.path.join() was (but with out the weird slash
        games), and can easily handle paths with a root directory other
        than /
    """
    _root = None
    _path = None

    def __new__(cls, path, *args, **kwds):
        root = kwds.pop("root", None)
        obj = str.__new__(cls, path, *args, **kwds)
        obj._path = path
        obj._root = None
        if root is not None:
            obj.newroot(str(root))
        return obj

    @property
    def ondisk(self):
        """ Path.ondisk evaluates as the real filesystem path of the path,
            including the path's root in the data.
        """
        if self.root:
            return normalize_path_slashes(Path(self.root) + Path(self.path))
        else:
            return Path(self.path)

    @property
    def path(self):
        return self._path

    @property
    def normpath(self):
        return Path(os.path.normpath(str(self.path)), root=self.root)

    @property
    def realpath(self):
        rp = os.path.realpath(self.ondisk)
        return Path(rp, root=self.root)

    @property
    def root(self):
        return self._root

    def newroot(self, newroot=None):
        """ Change the root directory of this Path """
        if newroot is None:
            self._root = None
        else:
            self._root = normalize_path_slashes(newroot)
            if self.startswith(self._root):
                path = self._path[len(self._root):]
                self._path = normalize_path_slashes(path)
        return self

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return repr(str(self.path))

    def __getitem__(self, idx):
        ret = str(self)
        return ret.__getitem__(idx)

    def __add__(self, other):
        if isinstance(other, Path):
            if other.root is not None and other.root != "/":
                if self.root is None:
                    self._root = other.root
                elif other.root != self.root:
                    raise ValueError("roots <%s> and <%s> don't match." %
                                     (self.root, other.root))
            path = "%s/%s" % (self.path, other.path)
        else:
            path = "%s/%s" % (self.path, other)
        path = normalize_path_slashes(path)
        return Path(path, root=self.root)

    def __eq__(self, other):
        if isinstance(other, Path):
            return self.path == other.path
        else:
            return self.path == str(other)

    def __lt__(self, other):
        if isinstance(other, Path):
            return self.path < other.path
        else:
            return self.path < str(other)

    def __gt__(self, other):
        if isinstance(other, Path):
            return self.path > other.path
        else:
            return self.path > str(other)

    def startswith(self, other):
        return self._path.startswith(str(other))

    def glob(self):
        """ Similar to glob.glob(), except it takes the Path's root into
            account when globbing and returns path objects with the same
            root, so you don't have to think about that part.
        """

        testdata_log.debug("glob: %s", self.ondisk)
        if "None" in self.ondisk:
            log.error("glob: %s", self.ondisk)
            log.error("^^ Somehow \"None\" got logged and that's never right.")
        for g in glob.glob(self.ondisk):
            testdata_log.debug("glob match: %s", g)
            yield Path(g, root=self.root)

    def __hash__(self):
        return self._path.__hash__()


def _run_program(argv, root='/', stdin=None, env_prune=None, stderr_to_stdout=False, binary_output=False):
    if env_prune is None:
        env_prune = []

    def chroot():
        if root and root != '/':
            os.chroot(root)

    with program_log_lock:  # pylint: disable=not-context-manager
        program_log.info("Running... %s", " ".join(argv))

        env = os.environ.copy()
        env.update({"LC_ALL": "C",
                    "INSTALL_PATH": root})
        for var in env_prune:
            env.pop(var, None)

        if stderr_to_stdout:
            stderr_dir = subprocess.STDOUT
        else:
            stderr_dir = subprocess.PIPE
        try:
            proc = subprocess.Popen(argv,  # pylint: disable=subprocess-popen-preexec-fn
                                    stdin=stdin,
                                    stdout=subprocess.PIPE,
                                    stderr=stderr_dir,
                                    close_fds=True,
                                    preexec_fn=chroot, cwd=root, env=env)

            out, err = proc.communicate()
            if not binary_output and six.PY3:
                out = out.decode("utf-8")
            if out:
                if not stderr_to_stdout:
                    program_log.info("stdout:")
                for line in out.splitlines():
                    program_log.info("%s", line)

            if not stderr_to_stdout and err:
                program_log.info("stderr:")
                for line in err.splitlines():
                    program_log.info("%s", line)

        except OSError as e:
            program_log.error("Error running %s: %s", argv[0], e.strerror)
            raise

        program_log.debug("Return code: %d", proc.returncode)

    return (proc.returncode, out)


def run_program(*args, **kwargs):
    return _run_program(*args, **kwargs)[0]


def capture_output(*args, **kwargs):
    return _run_program(*args, **kwargs)[1]


def capture_output_binary(*args, **kwargs):
    kwargs["binary_output"] = True
    return _run_program(*args, **kwargs)[1]


def run_program_and_capture_output(*args, **kwargs):
    return _run_program(*args, **kwargs)


def run_program_and_capture_output_binary(*args, **kwargs):
    kwargs["binary_output"] = True
    return _run_program(*args, **kwargs)


def mount(device, mountpoint, fstype, options=None):
    if options is None:
        options = "defaults"

    mountpoint = os.path.normpath(mountpoint)
    if not os.path.isdir(mountpoint):
        makedirs(mountpoint)

    argv = ["mount", "-t", fstype, "-o", options, device, mountpoint]
    return run_program(argv)


def umount(mountpoint):
    return run_program(["umount", mountpoint])


def get_mount_paths(dev):
    """ Given a device node path, return a list of all active mountpoints.

        :param str dev: Device path
        :returns: A list of mountpoints or []
        :rtype: list
    """
    from .mounts import mounts_cache

    mount_paths = mounts_cache.get_mountpoints(dev)
    if mount_paths:
        log.debug("%s is mounted on %s", dev, ', '.join(mount_paths))
    return mount_paths


def get_mount_device(mountpoint):
    """ Given a mountpoint, return the device node path mounted there. """
    mountpoint = os.path.realpath(mountpoint)  # eliminate symlinks
    mount_device = None
    with open("/proc/mounts") as mounts:
        for mnt in mounts.readline():
            try:
                (device, path, _rest) = mnt.split(None, 2)
            except ValueError:
                continue

            if path == mountpoint:
                mount_device = device
                break

    if mount_device and re.match(r'/dev/loop\d+$', mount_device):
        loop_name = os.path.basename(mount_device)
        mount_device = blockdev.loop.get_backing_file(loop_name)
        log.debug("found backing file %s for loop device %s", mount_device,
                  loop_name)

    if mount_device:
        log.debug("%s is mounted on %s", mount_device, mountpoint)

    return mount_device


def total_memory():
    """Return the amount of system RAM.

        Returns total system RAM in kB (given to us by /proc/meminfo).

        :rtype: :class:`~.size.Size`
    """
    # import locally to avoid a cycle with size importing util
    from .size import Size

    with open("/proc/meminfo", "r") as fobj:
        for line in fobj:
            if not line.startswith("MemTotal"):
                # we are only interested in the MemTotal: line
                continue

            fields = line.split()
            if len(fields) != 3:
                log.warning("unknown format for MemTotal line in /proc/meminfo: %s", line.rstrip())
                continue

            try:
                mem = Size("%s KiB" % fields[1])
            except ValueError:
                log.warning("invalid value of MemTotal /proc/meminfo: %s", fields[1])
                continue

            # Because /proc/meminfo only gives us the MemTotal (total physical RAM
            # minus the kernel binary code), we need to round this up. Assuming
            # every machine has the total RAM MiB number divisible by 128.
            bs = Size("128MiB")
            mem = (mem / bs + 1) * bs
            return mem

        raise RuntimeError("no valid line with MemTotal found in /proc/meminfo")


def available_memory():
    """ Return the amount of system RAM that is currenly available.

        :rtype: :class:`~.size.Size`
    """
    # import locally to avoid a cycle with size importing util
    from .size import Size

    with open("/proc/meminfo") as lines:
        mems = {k.strip(): v.strip() for k, v in (l.split(":", 1) for l in lines)}

    if "MemAvailable" in mems.keys():
        return Size("%s KiB" % mems["MemAvailable"].split()[0])
    else:
        # MemAvailable is not present on linux < 3.14
        free = Size("%s KiB" % mems["MemFree"].split()[0])
        cached = Size("%s KiB" % mems["Cached"].split()[0])
        buffers = Size("%s KiB" % mems["Buffers"].split()[0])

        return (free + cached + buffers)


##
# sysfs functions
##


def normalize_path_slashes(path):
    """ Normalize the slashes in a filesystem path.
        Does not actually examine the filesystme in any way.
    """
    while "//" in path:
        path = path.replace("//", "/")
    return path


def join_paths(*paths):
    """ Joins filesystem paths without any consiration of slashes or
        whatnot and then normalizes repeated slashes.
    """
    if len(paths) == 1 and hasattr(paths[0], "__iter__"):
        return join_paths(*paths[0])
    return normalize_path_slashes('/'.join(paths))


def get_sysfs_attr(path, attr, root=None):
    if not attr:
        log.debug("get_sysfs_attr() called with attr=None")
        return None
    if not isinstance(path, Path):
        path = Path(path=path, root=root)
    elif root is not None:
        path.newroot(root)

    attribute = path + attr
    fullattr = os.path.realpath(attribute.ondisk)

    if not os.path.isfile(fullattr) and not os.path.islink(fullattr):
        log.warning("%s is not a valid attribute", attr)
        return None

    f = open(fullattr, "r", encoding="utf-8", errors="replace")
    data = f.read()
    f.close()
    sdata = "".join(["%02x" % (ord(x),) for x in data])
    testdata_log.debug("sysfs attr %s: %s", attribute, sdata)
    return data.strip()


def sysfs_readlink(path, link, root=None):
    if not link:
        log.debug("sysfs_readlink() called with link=None")
    if isinstance(path, Path):
        linkpath = path + link
    else:
        linkpath = Path(path, root=root) + link

    linkpath = Path(os.path.normpath(linkpath), root=linkpath.root)
    fullpath = os.path.normpath(linkpath.ondisk)

    if not os.path.islink(fullpath):
        log.warning("%s is not a valid symlink", linkpath)
        return None

    output = os.readlink(fullpath)
    testdata_log.debug("new sysfs link: \"%s\" -> \"%s\"", linkpath, output)
    return output


def get_sysfs_path_by_name(dev_node, class_name="block"):
    """ Return sysfs path for a given device.

        For a device node (e.g. /dev/vda2) get the respective sysfs path
        (e.g. /sys/class/block/vda2). This also has to work for device nodes
        that are in a subdirectory of /dev like '/dev/cciss/c0d0p1'.
     """
    dev_name = os.path.basename(dev_node)
    if dev_node.startswith("/dev/"):
        dev_name = dev_node[5:].replace("/", "!")
    sysfs_class_dir = "/sys/class/%s" % class_name
    dev_path = os.path.join(sysfs_class_dir, dev_name)
    if os.path.exists(dev_path):
        return dev_path
    else:
        raise RuntimeError("get_sysfs_path_by_name: Could not find sysfs path "
                           "for '%s' (it is not at '%s')" % (dev_node, dev_path))


def get_path_by_sysfs_path(sysfs_path, dev_type="block"):
    """ Return device path for a given device sysfs path. """

    dev = get_sysfs_attr(sysfs_path, "dev")
    if not dev or not os.path.exists("/dev/%s/%s" % (dev_type, dev)):
        raise RuntimeError("get_path_by_sysfs_path: Could not find device for %s" % sysfs_path)
    return os.path.realpath("/dev/%s/%s" % (dev_type, dev))


def get_cow_sysfs_path(dev_path, dev_sysfsPath):
    """ Return sysfs path of cow device for a given device.
    """

    cow_path = dev_path + "-cow"
    if not os.path.islink(cow_path):
        raise RuntimeError("get_cow_sysfs_path: Could not find cow device for %s" % dev_path)

    # dev path for cow devices is actually a link to a dm device (e.g. /dev/dm-X)
    # we need the 'dm-X' name for sysfs_path (e.g. /sys/devices/virtual/block/dm-X)
    # where first part is the same as in sysfs_path of the original device
    dm_name = os.path.basename(os.path.realpath(cow_path))
    cow_sysfsPath = os.path.join(os.path.split(dev_sysfsPath)[0], dm_name)

    return cow_sysfsPath

##
# SELinux functions
##


def match_path_context(path, mode=0):
    """ Return the default SELinux context for the given path. """
    if not HAVE_SELINUX:
        raise RuntimeError("SELinux python bindings not available")
    context = None
    try:
        context = selinux.matchpathcon(os.path.normpath(path), mode)[1]
    except OSError as e:
        log.info("failed to get default SELinux context for %s: %s", path, e)

    return context


def set_file_context(path, context, root=None):
    """ Set the SELinux file context of a file.

        Arguments:

            path        filename string
            context     context string

        Keyword Arguments:

            root        an optional chroot string

        Return Value:

            True if successful, False if not.
    """
    if not HAVE_SELINUX:
        raise RuntimeError("SELinux python bindings not available")
    if root is None:
        root = '/'

    full_path = os.path.normpath("%s/%s" % (root, path))
    if context is None or not os.access(full_path, os.F_OK):
        return False

    try:
        rc = (selinux.lsetfilecon(full_path, context) == 0)
    except OSError as e:
        log.info("failed to set SELinux context for %s: %s", full_path, e)
        rc = False

    return rc


def reset_file_context(path, root=None, mode=0):
    """ Restore the SELinux context of a file to its default value.

        Arguments:

            path        filename string

        Keyword Arguments:

            root        an optional chroot string
            mode        an optional mode to use

        Return Value:

            If successful, returns the file's new/default context.
    """
    if not HAVE_SELINUX:
        raise RuntimeError("SELinux python bindings not available")
    context = match_path_context(path, mode)
    if context:
        if set_file_context(path, context, root=root):
            return context

##
# Miscellaneous
##


def makedirs(path):
    if not os.path.isdir(path):
        os.makedirs(path, 0o755)


def lsmod():
    """ Returns list of names of all loaded modules. """
    with open("/proc/modules") as f:
        lines = f.readlines()
    return [l.split()[0] for l in lines]


def get_option_value(opt_name, options):
    """ Return the value of a named option in the specified options string. """
    for opt in options.split(","):
        if "=" not in opt:
            continue

        name, val = opt.split("=")
        if name == opt_name:
            return val.strip()


def numeric_type(num):
    """ Verify that a value is given as a numeric data type.

        Return the number if the type is sensible or raise ValueError
        if not.
    """
    # import locally to avoid a cycle with size importing util
    from .size import Size

    if num is None:
        num = 0
    elif not isinstance(num, (six.integer_types, float, Size, Decimal)):
        raise ValueError("value (%s) must be either a number or None" % num)

    return num


def insert_colons(a_string):
    """ Insert colon between every second character.

        E.g. creates 'al:go:ri:th:ms' from 'algoritms'. Useful for formatting
        MAC addresses and wwids for output.
    """
    suffix = a_string[-2:]
    if len(a_string) > 2:
        return insert_colons(a_string[:-2]) + ':' + suffix
    else:
        return suffix


def sha256_file(filename):

    sha256 = hashlib.sha256()
    with open(filename, "rb") as f:

        block = f.read(65536)
        while block:
            sha256.update(block)
            block = f.read(65536)

    return sha256.hexdigest()


def read_file(filename, mode="r"):
    with open(filename, mode) as f:
        content = f.read()
    return content


class ObjectID(object):

    """This class is meant to be extended by other classes which require
       an ID which is preserved when an object copy is made.
       The value returned by the builtin function id() is not adequate:
       that value represents object identity so it is not in general
       preserved when the object is copied.

       The name of the identifier property is id, its type is int.

       The id is set during creation of the class instance to a new value
       which is unique for the object type. Subclasses can use self.id during
       __init__.
    """
    _newid_gen = functools.partial(next, itertools.count())

    def __new__(cls, *args, **kwargs):
        # pylint: disable=unused-argument
        self = super(ObjectID, cls).__new__(cls)
        self.id = self._newid_gen()  # pylint: disable=attribute-defined-outside-init,assignment-from-no-return
        return self


def canonicalize_UUID(a_uuid):
    """ Converts uuids to canonical form.

        :param str a_uuid: the UUID

        :returns: a canonicalized UUID
        :rtype: str

        mdadm's UUIDs are actual 128 bit uuids, but it formats them strangely.
        This converts the uuids to canonical form.
        Example:
            mdadm UUID: '3386ff85:f5012621:4a435f06:1eb47236'
        canonical UUID: '3386ff85-f501-2621-4a43-5f061eb47236'

        If the UUID is already in canonical form, the conversion
        is equivalent to the identity.
    """
    return str(uuid.UUID(a_uuid.replace(':', '')))

# Most Python 2/3 compatibility code equates python 2 str with python 3 bytes,
# but the equivalence that we actually need to avoid return type surprises is
# str/str.


def stringize(inputstr):
    """ Convert strings to a format compatible with Python 2's str.

        :param str inputstr: the string to convert

        :returns: a string with the correct type
        :rtype: str

        This method is for use in __str__ calls to ensure that they always
        return a str. In Python 3, this method simply inputstr as a string. In
        Python 2, it converts unicode into str. The returned str in python 2 is
        encoded using utf-8.
    """
    if six.PY2:
        # pylint: disable=undefined-variable
        if isinstance(inputstr, unicode):
            inputstr = inputstr.encode('utf-8')

    return str(inputstr)

# Like six.u, but without the part where it raises an exception on unicode
# objects


def unicodeize(inputstr):
    """ Convert strings to a format compatible with Python 2's unicode.

        :param str inputstr: the string to convert

        :returns: a string with the correct type
        :rtype: unicode

        This method is for use in __unicode__ calls to ensure that they always
        return a unicode. This method does not handle non-ASCII characters
        in str parameters, but non-ASCII characters in unicode parameters will
        be correctly passed through.
    """
    if six.PY2:
        # pylint: disable=undefined-variable
        return unicode(inputstr)
    else:
        return str(inputstr)


def compare(first, second):
    """ Compare two objects.

        :param first: first object to compare
        :param second: second object to compare
        :returns: 0 if first == second, 1 if first > second, -1 if first < second
        :rtype: int

        This method replaces Python 2 cmp() built-in-function.
    """

    if first is None and second is None:
        return 0

    elif first is None:
        return -1

    elif second is None:
        return 1

    else:
        return (first > second) - (first < second)


def dedup_list(alist):
    """Deduplicates the given list by removing duplicates while preserving the order"""
    seen = set()
    ret = []
    for item in alist:
        if item not in seen:
            ret.append(item)
        seen.add(item)
    return ret


##
# Convenience functions for examples and tests
##


def set_up_logging(log_dir="/tmp", log_prefix="blivet", console_logs=None):
    """ Configure the blivet logger to write out a log file.

        :keyword str log_dir: path to directory where log files are
        :keyword str log_prefix: prefix for log file names
        :keyword list console_logs: list of log names to output on the console
    """
    log.setLevel(logging.DEBUG)
    program_log.setLevel(logging.DEBUG)

    def make_handler(path, prefix, level):
        log_file = "%s/%s.log" % (path, prefix)
        log_file = os.path.realpath(log_file)
        handler = logging.FileHandler(log_file)
        handler.setLevel(level)
        formatter = logging.Formatter("%(asctime)s %(levelname)s %(name)s/%(threadName)s: %(message)s")
        handler.setFormatter(formatter)
        return handler

    handler = make_handler(log_dir, log_prefix, logging.DEBUG)
    log.addHandler(handler)
    program_log.addHandler(handler)

    # capture python warnings in our logs
    warning_log = logging.getLogger("py.warnings")
    warning_log.addHandler(handler)

    if console_logs:
        set_up_console_log(log_names=console_logs)

    log.info("sys.argv = %s", sys.argv)

    prefix = "%s-testdata" % (log_prefix,)
    handler = make_handler(log_dir, prefix, logging.DEBUG)
    testdata_log.setLevel(logging.DEBUG)
    testdata_log.addHandler(handler)


def set_up_console_log(log_names=None):
    log_names = log_names or []
    handler = logging.StreamHandler()
    console_log.setLevel(logging.DEBUG)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(threadName)s: %(message)s")
    handler.setFormatter(formatter)
    console_log.addHandler(handler)
    for name in log_names:
        logging.getLogger(name).addHandler(handler)


def create_sparse_tempfile(name, size):
    """ Create a temporary sparse file.

        :param str name: suffix for filename
        :param :class:`~.size.Size` size: the file size
        :returns: the path to the newly created file
    """
    (fd, path) = tempfile.mkstemp(prefix="blivet.", suffix="-%s" % name)
    os.close(fd)
    create_sparse_file(path, size)
    return path


def create_sparse_file(path, size):
    """ Create a sparse file.

        :param str path: the full path to the file
        :param :class:`~.size.Size` size: the size of the file
        :returns: None
    """
    fd = os.open(path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC)
    os.ftruncate(fd, size)
    os.close(fd)


@contextmanager
def sparsetmpfile(name, size):
    """ Context manager that creates a sparse tempfile and then unlinks it.

        :param str name: suffix for filename
        :param :class:`~.size.Size` size: the file size

        Yields the path to the newly created file on __enter__.
    """
    path = create_sparse_tempfile(name, size)
    try:
        yield path
    finally:
        os.unlink(path)


def variable_copy(obj, memo, omit=None, shallow=None, duplicate=None):
    """ A configurable copy function. Any attributes not specified in omit,
        shallow, or duplicate are copied using copy.deepcopy().

        :param object obj: a python object to be copied.
        :param dict memo: a dictionary of already copied items
        :param omit: a list of names of attributes not to copy
        :type omit: iterable of str
        :param shallow: a list of names of attributes to shallow copy
        :type shallow: iterable of str
        :param duplicate: a list of names of attributes to duplicate
        :type duplicate: iterable of str

        Note that all atrributes in duplicate must implement a duplicate()
        method that does what is expected of it. Attributes with type
        pyparted.Disk are known to do so.

        A shallow copy is implemented by calling copy.copy().
    """
    omit = omit or []
    shallow = shallow or []
    duplicate = duplicate or []

    new = obj.__class__.__new__(obj.__class__)
    memo[id(obj)] = new
    for (attr, value) in obj.__dict__.items():
        if attr in omit or value is None:
            setattr(new, attr, value)
        elif attr in shallow:
            setattr(new, attr, copy.copy(value))
        elif attr in duplicate:
            setattr(new, attr, value.duplicate())
        else:
            setattr(new, attr, copy.deepcopy(value, memo))

    return new


def get_current_entropy():
    with open("/proc/sys/kernel/random/entropy_avail", "r") as fobj:
        return int(fobj.readline())


def power_of_two(value):
    """ Checks whether value is a power of 2 greater than 1.

        :param any value: a value
        :returns: True if the value is a power of 2
        :rtype: bool
    """
    try:
        int_value = int(value)
    except (ValueError, TypeError):
        return False

    if int_value != value:
        return False

    value = int_value

    if value < 2:
        return False

    (q, r) = (value, 0)
    while q != 0:
        if r != 0:
            return False
        (q, r) = divmod(q, 2)

    return True


def indent(text, spaces=4):
    """ Indent text by a specified number of spaces.

        :param str text: the text to indent
        :keyword int spaces: the number of spaces to indent text

        It would be nice if we could use textwrap.indent for this but, since it
        does not exist in python2, I prefer to just use this.
    """
    if not text or not text.strip():
        return text

    indentation = " " * spaces
    indented = []
    for line in text.splitlines():
        indented.append("%s%s" % (indentation, line))

    return "\n".join(indented)


def _add_extra_doc_text(func, field=None, desc=None, field_unique=False):
    """ Add extra doc text to a function's docstring.

        :param :class:`function` func: the function
        :param str field: (sphinx) field to add to the doc text
        :param str desc: description to add in the given :param:`field`
        :param bool field_unique: whether the given :param:`field` should only
                                  appear in the doc text once (a new one won't
                                  be added if there already is an existing one)

        If your doctext is indented with something other than spaces the added
        doctext's indentation will probably not match. That'd be your fault.
    """

    base_text = func.__doc__
    if base_text is None:
        base_text = " "  # They contain leading and trailing spaces. *shrug*
    else:
        base_text = base_text[:-1]  # Trim the trailing space.

    if field_unique and field in base_text:
        # Don't add multiple fields
        return

    # Figure out the number of spaces to indent docstring text. We are looking
    # for the minimum indentation, not including the first line or empty lines.
    indent_spaces = None
    for l in base_text.splitlines()[1:]:
        if not l.strip():
            continue

        spaces = 0
        _l = l[:]
        while _l and _l.startswith(" "):
            spaces += 1
            _l = _l[1:]

        if indent_spaces is None or indent_spaces > spaces:
            indent_spaces = spaces

    if indent_spaces is None:
        indent_spaces = 0

    text = ""
    if not re.search(r'\n\s*$', base_text):
        # Make sure there's a newline after the last text.
        text = "\n"

    desc = desc or ""
    text += field + " " + desc
    func.__doc__ = base_text + "\n" + indent(text, indent_spaces)


#
# Deprecation decorator.
#
_DEPRECATION_MESSAGE = "will be removed in a future version."


def _default_deprecation_msg(func):
    return "%s %s" % (func.__name__, _DEPRECATION_MESSAGE)


_SPHINX_DEPRECATE = ".. deprecated::"
_DEPRECATION_INFO = """%(version)s
    %(message)s
"""


def _add_deprecation_doc_text(func, version=None, message=None):
    desc = _DEPRECATION_INFO % {"version": version, "message": message}
    _add_extra_doc_text(func, _SPHINX_DEPRECATE, desc, field_unique=True)


def deprecated(version, message):
    """ Decorator to deprecate a function or method via warning and docstring.

        :param str version: version in which the deprecation is effective
        :param str message: message suggesting a preferred alternative

        .. note::
            At the point this decorator gets applied to a method in a class the
            method is just a function. It becomes a method later.

        The docstring manipulation is performed only once for each decorated
        function/method, but the warning is issued every time the decorated
        function is called.
    """
    def deprecate_func(func):
        @wraps(func)
        def the_func(*args, **kwargs):
            """ Issue a deprecation warning for, then call, a function. """
            # Warnings look much better with default warning text than with
            # no text. The sphinx doesn't benefit from it, so don't use it
            # there.
            warn_msg = _default_deprecation_msg(func)
            if message:
                warn_msg += " %s" % message

            warnings.warn(warn_msg, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        _add_deprecation_doc_text(the_func, message=message, version=version)
        return the_func

    return deprecate_func


def default_namedtuple(name, fields, doc=""):
    """Create a namedtuple class

    The difference between a namedtuple class and this class is that default
    values may be specified for fields and fields with missing values on
    initialization being initialized to None.

    :param str name: name of the new class
    :param fields: field descriptions - an iterable of either "name" or ("name", default_value)
    :type fields: list of str or (str, object) objects
    :param str doc: the docstring for the new class (should at least describe the meanings and
                    types of fields)
    :returns: a new default namedtuple class
    :rtype: type

    """
    field_names = list()
    for field in fields:
        if isinstance(field, tuple):
            field_names.append(field[0])
        else:
            field_names.append(field)
    nt = namedtuple(name, field_names)

    class TheDefaultNamedTuple(nt):
        if doc:
            __doc__ = doc

        def __new__(cls, *args, **kwargs):
            args_list = list(args)
            sorted_kwargs = sorted(kwargs.keys(), key=field_names.index)
            for i in range(len(args), len(field_names)):
                if field_names[i] in sorted_kwargs:
                    args_list.append(kwargs[field_names[i]])
                elif isinstance(fields[i], tuple):
                    args_list.append(fields[i][1])
                else:
                    args_list.append(None)

            return nt.__new__(cls, *args_list)

    return TheDefaultNamedTuple


def requires_property(prop_name, val=True):
    """
    Function returning a decorator that can be used to guard methods and
    properties with evaluation of the given property.

    :param str prop_name: property to evaluate
    :param val: guard value of the :param:`prop_name`
    :type val: :class:`Object` (anything)
    """
    def guard(fn):
        @wraps(fn)
        def func(self, *args, **kwargs):
            if getattr(self, prop_name) == val:
                return fn(self, *args, **kwargs)
            else:
                raise ValueError("%s can only be accessed if %s evaluates to %s" % (fn.__name__, prop_name, val))
        return func
    return guard


class EvalMode(Enum):
    onetime = 1
    always = 2
    # TODO: no_sooner_than, if_changed,...


@six.add_metaclass(abc.ABCMeta)
class DependencyGuard(object):

    error_msg = abc.abstractproperty(doc="Error message to report when a dependency is missing")

    def __init__(self, exn_cls=DependencyError):
        self._exn_cls = exn_cls
        self._avail = None

    def check_avail(self, onetime=False):
        if self._avail is None or not onetime:
            self._avail = self._check_avail()
        return self._avail

    @abc.abstractmethod
    def _check_avail(self):
        raise NotImplementedError()

    def __call__(self, critical=False, eval_mode=EvalMode.always):
        def decorator(fn):
            @wraps(fn)
            def decorated(*args, **kwargs):
                just_onetime = eval_mode == EvalMode.onetime
                if self.check_avail(onetime=just_onetime):
                    return fn(*args, **kwargs)
                elif critical:
                    raise self._exn_cls(self.error_msg)
                else:
                    log.warning("Failed to call the %s method: %s", fn.__name__, self.error_msg)
                    return None
            return decorated
        return decorator


def detect_virt():
    """ Return True if we are running in a virtual machine. """
    try:
        vm = safe_dbus.get_property_sync(SYSTEMD_SERVICE, SYSTEMD_MANAGER_PATH,
                                         SYSTEMD_MANAGER_IFACE, VIRT_PROP_NAME)
    except (safe_dbus.DBusCallError, safe_dbus.DBusPropertyError):
        return False
    else:
        return vm[0] in ('qemu', 'kvm', 'xen')


def natural_sort_key(device):
    """ Sorting key for devices which makes sure partitions are sorted in natural
        way, e.g. 'sda1, sda2, ..., sda10' and not like 'sda1, sda10, sda2, ...'
    """
    if device.type == "partition" and device.parted_partition and device.disk:
        part_num = getattr(device.parted_partition, "number", -1)
        return [device.disk.name, part_num]
    else:
        return [device.name, 0]


def get_kernel_module_parameter(module, parameter):
    """ Return the value of a given kernel module parameter

    :param str module: a kernel module
    :param str parameter: a module parameter
    :returns: the value of the given kernel module parameter or None
    :rtype: str
    """

    value = None

    parameter_path = os.path.join("/sys/module", module, "parameters", parameter)
    try:
        with open(parameter_path) as f:
            value = f.read().strip()
    except IOError as e:
        log.warning("Couldn't get the value of the parameter '%s' from the kernel module '%s': %s",
                    parameter, module, str(e))

    return value
